"""
Code Generator for Smart Home - The Next Generation.

Generates helper code from component manifests.

Smart Home - TNG is a Home Automation framework for observing the state
of entities and react to changes. It is based on Home Assistant from
home-assistant.io and the Home Assistant Community.

Copyright (c) 2022, Andreas Nixdorf

This program is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program.  If not, see
http://www.gnu.org/licenses/.
"""

import pprint
import re
import typing

from .code_generator import CodeGenerator
from .config import Config
from .integration import Integration

_NAME: typing.Final = "dhcp"
_BASE: typing.Final = """
\"\"\"Automatically generated by shc_from_manifests.

To update, run python3 -m script.shc_from_manifests.
\"\"\"

import typing

# fmt: off

# pylint: disable=unused-variable
DHCP: typing.Final[list[dict[str, str | bool]]] = {}
""".strip()


# pylint: disable=unused-variable
class DhcpGenerator(CodeGenerator):
    """Generate dhcp file."""

    def __init__(self):
        super().__init__(_NAME)

    def generate_and_validate(self, integrations: list[dict[str, str]], config: Config):
        """Validate and generate dhcp data."""
        match_list = []

        for domain in sorted(integrations):
            integration = integrations[domain]

            if not integration.manifest or not integration.config_flow:
                continue

            match_types = integration.manifest.get("dhcp", [])

            if not match_types:
                continue

            for entry in match_types:
                match_list.append({"domain": domain, **entry})

        # JSON will format `True` as `true`
        # re.sub for flake8 E128
        formatted = pprint.pformat(match_list)
        formatted_aligned_continuation = re.sub(r"^\[\{", "[\n {", formatted)
        formatted_align_indent = re.sub(
            r"(?m)^ ",
            "    ",
            formatted_aligned_continuation,
            flags=re.MULTILINE,
            count=0,
        )
        return _BASE.format(formatted_align_indent)

    def validate(self, integrations: dict[str, Integration], config: Config):
        """Validate dhcp file."""
        dhcp_path = config.root / "smart_home_tng/core/generated/dhcp.py"
        config.cache["dhcp"] = content = self.generate_and_validate(integrations, config)

        if config.specific_integrations:
            return

        if not dhcp_path.is_file():
            config.add_error(
                "dhcp",
                "File dhcp.py is not up to date. Run python3 -m script.shc_from_manifests.",
                fixable=True,
            )
            return

        with open(str(dhcp_path), encoding="utf-8") as fp:
            current = fp.read().strip()
            if current != content:
                config.add_error(
                    "dhcp",
                    "File dhcp.py is not up to date. Run python3 -m script.shc_from_manifests.",
                    fixable=True,
                )
            return

    def generate(self, integrations: dict[str, Integration], config: Config):
        """Generate dhcp file."""
        dhcp_path = config.root / "smart_home_tng/core/generated/dhcp.py"
        with open(str(dhcp_path), "w", encoding="utf-8") as fp:
            fp.write(f"{config.cache['dhcp']}\n")
